/*****************************************************************************/
//  cf_callback.hh
//  Copyright (C) 1999 Brett Hall <swizin@rain.org>
//  
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or 
//  (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 USA 
/*****************************************************************************/


//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// cf_callback.hh
//	-CF_CallBack class: used to define "built_in" functions
//
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

#ifndef CF_CALLBACK_HH
#define CF_CALLBACK_HH

#include <string>
#include <vector>
#include <map>

using namespace std;

//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//CF_CallBack

/**Base class for "built-in" functions.
	All built-in functions should sub-class this class and redefine Call() in order to implement
	their own functionality
*/
class CF_CallBack
{
public:
	/**Creates a callback function
		@param nameIn Name for function (used to call it)
		@param minArgs Minimum number of arsg the command requires, only used during script parsing
	*/
	CF_CallBack(const string& nameIn, unsigned int minArgsIn = 0): name(nameIn), minArgs(minArgsIn) {}
	
	//Virtual destructor, does nothing
	virtual ~CF_CallBack() {}
	
	/**Sub class should redefine this, called when command is called
		@param args Arguments to command
		@return Return value for function
	*/
	virtual string Call(vector<string>& args) = 0;
	
	/**Gets the commands name
		@return commands name
	*/
	const string& GetName() {return name;}
	
	/**Gets the commands minimum number of arguments
		@return Minimum number of args
	*/
	unsigned int GetMinArgs() {return minArgs;}
	
protected:
	string name;
	unsigned int minArgs;
};

//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// class CF_CommandMgr
//	-manages built in commands
//	-can only add new functions before mgr is registered with a widget (to avoid name collisions with script variables)

/**The command manager.
	This class manages "built-in" commands.  All commands must be registered using
	RegisterFunction() before the manager is attched to any control freak widgets.  Once the 
	manager is "in use" commands cannot be added.
*/
class CF_CommandMgr 
{
public:
	CF_CommandMgr()	{ inUse = false;}
	
	/**Registers a function for use
		@param function pointer to callback to register
		@return True if no name collision with existing commands
	*/
	bool RegisterFunction(CF_CallBack* function);
	
	/**Calls the given command
		@param name Name of command to call
		@param args Arguments for command
		@return Return value of command
	*/
	string CallFunction(const string& name, vector<string>& args);
	
	///Sets the manager's "in use" flag, the flag cannot be turned off once it is on
	void SetInUse() {inUse = true;}
	
	///Checks for command existence
	bool CheckFunction(const string& name) {return functions[name] != NULL;}
	
private:
	map<const string, CF_CallBack*> functions;
	bool inUse;
};//class CF_CommandMgr

//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// class CF_Option

/**Client option class.  The client program that is creating a control freak widget
	can register options that the widget will search for when parsing the script file.  	
	The client uses the CF_Option class to communicate what options it is looking for to
	the control freak widget.  The options that the client is looking for must be registered
	with the CF_OptionMgr passed to the control freak widget when it is created.  After the
	script file has been parsed the results of the option search can be obtained by querying
	the option mgr used.
*/
class CF_Option 
{
public:
	/**Creates an option (don't use directly, use CF_OptionMgr::AddOption() instead)
		@param nameIn Name for option
	*/
	CF_Option(const string& nameIn): found(false) {}
	
	///Returns true if the option was found
	bool Found()	{return found;}
	
	/**Called to set the option's "found" flag
		@param argIn The argument found for the option
	*/
	void SetFound(const string& argIn)	{arg = argIn; found = true;}
	
	///Gets the argument found for the option
	string GetArg()	{return arg;}	
	
private:
	string arg;
	bool found;
};//class CF_Option


//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// class CF_OptionsMgr
//	-manages app specific options in script file

/**Client option manager.  Manages the options that the client is looking for in a script file.
	The client program can take options from the script file that is used to create a control
	freak widget.  These options are registered using AddOption().  The results of the search can 
	be obtained by using QueryOption() and the argument for the option gotten from GetArg().
*/ 
class CF_OptionMgr 
{
public:
	/**Registers an option to search for
		@param name Name for option
		@return True if no name collision with an existing option
	*/
	bool AddOption(const string& name);
	
	/**Queries whether an option was found or not
		@param name Name of option to query
		@return True if option was found in script file
	*/
	bool QueryOption(const string& name);
	
	/**Get the argument found for an option
		@param name Name of option to get argument for
		@return Arguement found for option, empty string if option wasn't found (use QueryOption()
			to determine if option was found, found options can have empty arguments as well).
	*/
	string GetArg(const string& name);
	
	/**Sets an option's "found" flag
		@param name Name of option found
		@param arg Argument found for option
		@return True if option exists
	*/
	bool SetOptionFound(const string& name, const string& arg);
	
	//Deletes all options
	~CF_OptionMgr();
	
private:
	map<const string, CF_Option*> options;
};//class CF_OptionsMgr


#endif
