/*****************************************************************************/
//  cf_command.cc
//  Copyright (C) 1999 Brett Hall <swizin@rain.org>
//  
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or 
//  (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 USA 
/*****************************************************************************/


//////////////////////////////////////////////////
// cf_command.cc
//
//////////////////////////////////////////////////


#include "cf_command.hh"
#include "cf_exception.hh"
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stack>
#include <gtk--/main.h>

///////////////////////////////////////////////////
//forward eclarations
static bool IsFunction(const string& cmdStr);

//////////////////////////////////////////////////
//BEGIN CF_CommandElement

CF_VarMgr* CF_CommandElement::curVarMgr = NULL;
CF_CommandMgr* CF_CommandElement::curCmdMgr = NULL;

//CF_CommandElement::CreateCommandElement: parses string and creates top level command elements
//	cmdStr: string to parse
//	cmdMgr: command manager to get built in cmds from
//	varMgr: script var mgr to get script vars from
//returns:top level command element
CF_CommandElement* CF_CommandElement::CreateCommandElement(const string& cmdStr, CF_CommandMgr* cmdMgr, CF_VarMgr* varMgr)
{
	//value to return:
	CF_CommandElement* ce = NULL;
	
	//update current command and var managers
	curCmdMgr = cmdMgr;
	curVarMgr = varMgr;
	
	//check for shell or built-in command
	if(IsFunction(cmdStr))
		ce = reinterpret_cast<CF_CommandElement*>(new CF_CEBuiltIn(cmdStr));
	else
		ce = reinterpret_cast<CF_CommandElement*>(new CF_CEShell(cmdStr));
	CF_Assert(ce != NULL, "CF_CommandElement::CreateCommandElement()");
	
	return ce;
}//CF_CommandElement* CF_CommandElement::CreateCommandElement(string cmdStr, CF_CommandMgr* cmdMgr, CF_VarMgr* varMgr)

void CF_CommandElement::CreateCommands(const string& cmdStr, CEList& cmds,
										 CF_CommandMgr* functions, CF_VarMgr* varMgr)
{
	//break into separate cmd strings and create cmds
	int depth = 0;
	string curCmd;
	for(size_t pos = 0; pos < cmdStr.length(); ++pos)
	{
		char cur = cmdStr[pos];
		switch(cur)
		{
		case ';':
			if(depth == 0)
			{
				cmds.AddCommand(CF_CommandElement::CreateCommandElement(curCmd, functions, varMgr));
				curCmd = "";			
			}//if(depth == 0)
			else
			{
				curCmd += cur;
			}//else
			break;
		
		case '{':
		case '(':
			++depth;
			curCmd += cur;
			break;
		
		case '}':
		case ')':
			--depth;
			curCmd += cur;
			break;
			
		default:
			curCmd += cur;
			break;
		}//switch(cur)
	}//for(int pos = 0; pos < cmdStr.length(); ++pos) 
	//make sure final cmd is created
	if(!curCmd.empty())
		cmds.AddCommand(CF_CommandElement::CreateCommandElement(curCmd, functions, varMgr));
}//void CF_ToolbarWidget::CreateCommands(const string& cmdStr, list<CF_CommandElement*>& cmdList)

static bool IsControlChar(char cur, char prev)
{
	//cheeck for escpae character
	if(prev == '\\')
		return false;
		
	switch(cur)
	{
	case ',':
	case '!':
	case '{':
	case '}':
	case '(':
	case ')':
	case '?':
	case ':':
		return true;
		
	default:
		return false;
	}//switch(cur)
}//static bool IsControlChar(char cur, char prev)

enum PARSE_STATE {STRING_EL, EXPAND_EL, FUNCTION_EL, SHELL_EL, DECISION_EL};

CF_CommandElement* CF_CommandElement::CreateCommandElement(const string& cmdStr)
{
	//make sure command string isn't empty
	if(cmdStr.empty())
	{
		CF_CommandElement* newElement = reinterpret_cast<CF_CommandElement*>(new CF_CEString(cmdStr));
		CF_Assert(newElement != NULL, "CF_CommandElement::CreateCommadnLement()");
		return newElement;
	}//if(cmdStr.empty())
	
	//init for parsing
	stack<PARSE_STATE> state;
	state.push(STRING_EL);
	int depth = 0;
	string curElement;
	char prevChar = ' ';
	list<pair<string, Type> > elements;
	
	//parse string to build element list
	for(size_t pos = 0; pos < cmdStr.length(); ++pos)
	{
		//check for control character
		char curChar = cmdStr[pos];
		if(IsControlChar(curChar, prevChar))
		{
			switch(state.top())
			{
			case STRING_EL:
				switch(curChar)
				{
				case '(':
					state.push(FUNCTION_EL);
					curElement += curChar;
					break;
					
				case '{':
					if(!curElement.empty())
					{
						elements.push_back(pair<string, Type>(curElement, STRING));
						curElement.erase();	
					}//if(!curElement.empty())
					if(cmdStr[pos + 1] == '!')
					{
						pos++;
						state.push(SHELL_EL);
					}//if(cmdStr[pos + 1] == '!')
					else
					{
						state.push(EXPAND_EL);
					}//else
					break;
										
				case '!':
				case ')':
				case '}':
				case ',':
				case '?':
				case ':':		
				default:
					throw CF_ParseException("CF_CommandElement::CreateCommandElement()", "Syntax error in command");
				}//switch(curChar)
				break;
				
			case EXPAND_EL:
				if(depth == 0)
				{
					switch(curChar)
					{
					case '}':
						if(!curElement.empty())
						{
							elements.push_back(pair<string, Type>(curElement, VARIABLE));
							curElement.erase();						
						}//if(!curElement.empty())
						state.pop();
						break;
						
					case '{':
						++depth;
						curElement += curChar;
						break;
						
					case '(':
						state.push(FUNCTION_EL);
						curElement += curChar;
						break;
						
					case '?':
						state.top() = DECISION_EL;
						curElement += curChar;
						break;
					
					case ')':
					case '!':
					case ',':
					case ':':
					default:
						throw CF_ParseException("CF_CommandElement::CreateCommandElement()", "Syntax error in command");
					}//switch(curChar)
				}//if(depth == 0)
				else
				{
					switch(curChar)
					{
					case '{':
						++depth;
						break;
						
					case '}':
						--depth;
						break;
						
					default:
						break;
					}//switch(curChar)
					curElement += curChar;
				}//else
				break;
				
			case FUNCTION_EL:
				if(depth == 0)
				{
					switch(curChar)
					{
					case '{':
						++depth;
						curElement += curChar;
						break;
						
					case ')':
						curElement += curChar;
						elements.push_back(pair<string, Type>(curElement, BUILTIN));
						curElement.erase();
						state.pop();
						if(state.top() == STRING_EL)
						{
							if(pos != (cmdStr.length() - 1))
								throw CF_ParseException("CF_CommandElement::CreateCommandElement()", "Syntax error in command");
						}//if(state.top() == STRING_EL)
						else if(state.top() == EXPAND_EL)
						{
							if(cmdStr[pos + 1] != '}')
								throw CF_ParseException("CF_CommandElement::CreateCommandElement()", "Syntax error in command");							
						}//else if(state.top() == EXPAND_EL)
						break;
						
					case ',':
						curElement += curChar;
						break;
						
					case '}':
					case '(':
					case '!':
					case '?':
					case ':':
					default:
						throw CF_ParseException("CF_CommandElement::CreateCommandElement()", "Syntax error in command");	
					}//switch(curChar)
				}//if(depth == 0)
				else
				{
					switch(curChar)
					{
					case '{':
						++depth;
						break;
						
					case '}':
						--depth;
						break;
						
					default:
						break;
					}//switch(curChar)
					curElement += curChar;
				}//else
				break;
				
			case SHELL_EL:
				if(depth == 0)
				{
					switch(curChar)
					{
					case '{':
						++depth;
						curElement += curChar;
						break;
						
					case '}':
						elements.push_back(pair<string, Type>(curElement, SHELL));
						curElement.erase();
						state.pop();
						break;
						
					case '!':
					case ',':
					case '(':
					case ')':
					case '?':
					case ':':
					default:
						throw CF_ParseException("CF_CommandElement::CreateCommandElement()", "Syntax error in command");	
					}//switch(curChar)
				}//if(depth == 0)
				else
				{
					switch(curChar)
					{
					case '{':
						++depth;
						break;
						
					case '}':
						--depth;
						break;
						
					default:
						break;
					}//switch(curChar)
					curElement += curChar;
				}//else
				break;
				
			case DECISION_EL:
				if(depth == 0)
				{
					switch(curChar)
					{
					case '{':
						++depth;
						curElement += curChar;
						break;

					case '?':
					case ':':
						curElement += curChar;
						break;

					case '}':
						elements.push_back(pair<string, Type>(curElement, DECISION));
						curElement.erase();
						state.pop();
						break;

					case ',':
					case '!':
					case '(':
					case ')':
					default:
						throw CF_ParseException("CF_CommandElement::CreateCommandElement()", "Syntax error in command");							
					}//switch(curChar)
				}//if(depth == 0)
				else
				{
					switch(curChar)
					{
					case '{':
						++depth;
						break;
						
					case '}':
						--depth;
						break;
						
					default:
						break;
					}//switch(curChar)
					curElement += curChar;
				}//else
			}//switch(state.top())
		}//if(IsControlChar(curChar, prevChar))
		else
		{
			//not a control char so add to element
			curElement += curChar;
		}//else
		
		//update previous character
		prevChar = curChar;
	}//for(size_t pos = 0; pos < cmdStr.length(); ++pos)
	
	//check for errors
	if((depth != 0) || (state.top() != STRING_EL) || (state.size() != 1))
		throw CF_ParseException("CF_CommandElement::CreateCommandElement()", "Syntax error in command");	
		
	//push last element
	if(!curElement.empty())
		elements.push_back(pair<string, Type>(curElement, STRING));
		
	//last error check	
	if(elements.empty())
		throw CF_ParseException("CF_CommandElement::CreateCommandElement()", "Syntax error in command");
		
	//check list length and create element
	CF_CommandElement* newElement = NULL;
	if(elements.size() == 1)
	{
		switch(elements.front().second)
		{
		case STRING:
			newElement = reinterpret_cast<CF_CommandElement*>(new CF_CEString(elements.front().first));
			break;
			
		case SHELL:
			newElement = reinterpret_cast<CF_CommandElement*>(new CF_CEShell(elements.front().first));
			break;
			
		case BUILTIN:
			newElement = reinterpret_cast<CF_CommandElement*>(new CF_CEBuiltIn(elements.front().first));
			break;
			
		case VARIABLE:
			newElement = reinterpret_cast<CF_CommandElement*>(new CF_CEVariable(elements.front().first));
			break;
			
		case DECISION:
			newElement = reinterpret_cast<CF_CommandElement*>(new CF_CEDecision(elements.front().first));
			break;
		
		case LIST:
			throw CF_ParseException("CF_CommandElement::CreateCommandElement()", "Syntax error in command");
			break;
		}//switch(elemnts.front().second)
	}//if(elements.size() == 1)
	else
	{
		newElement = reinterpret_cast<CF_CommandElement*>(new CF_CEList(elements));
	}//else
	CF_Assert(newElement != NULL, "CF_CommandElement::CreateCommandElement()");
	return newElement;
}//CF_CommandElement* CF_CommandElement::CreateCommandElement(string cmdStr)


//END CF_CommandElement


/////////////////////////////////////////////////////
//BEGIN CF_CEString

CF_CEString::CF_CEString(const string& cmdStr)
{
	//set type
	type = STRING;
	
	//get rid of conrtol characters
	char prevChar = ' ';
	for(size_t pos = 0; pos < cmdStr.length(); ++pos)
	{
		char curChar = cmdStr[pos];
		if((curChar != '\\') || (prevChar == '\\'))
			str += curChar;
		prevChar = curChar;	
	}//for(size_t pos = 0; pos < cmdStr.length(); ++pos)
}//CF_CEString::CF_CEString(const string& cmdStr)

//END CF_CEString


/////////////////////////////////////////////////////
//BEGIN CF_CEVariable

//CF_CEVariable::CF_CEVariable: creates a variable command element
//	cmdStr: string to parse as var name
//returns: <nothing>
CF_CEVariable::CF_CEVariable(const string& cmdStr)
{
	//set type
	type = CF_CommandElement::VARIABLE;
	
	//create command name command element
	varName = CreateCommandElement(cmdStr);
	
	//save var mgr
	varMgr = curVarMgr;
}//CF_CEVariable::CF_CEVariable(string cmdStr)

//CF_CEVariable::~CF_CEVariable: gets rid of varName element
//returns: <nothing>
CF_CEVariable::~CF_CEVariable()
{
	if(varName != NULL)
		delete varName;
}//CF_CEVariable::~CF_CEVariable()

//CF_CEVariable::Expand: Expands variable
//returns: value of variable
string CF_CEVariable::Expand()
{
	string name = varName->Expand();
	CF_ScriptVar* var =  varMgr->GetVar(name);
	if(var == NULL)
	{
		cerr << "Bad variable name: " << name << endl;
		return "";
	}//if(var == NULL)
	return var->GetStringValue();
}//string CF_CEVariable::Expand()


//END CF_CEVariable


/////////////////////////////////////////////////////
//BEGIN CF_CEDecision

//CF_CEDecision::CF_CEDecision: creates a decision command element
//	cmd: string to parse
//returns: <nothing>
CF_CEDecision::CF_CEDecision(const string& cmd)
{
	//set type
	type = CF_CommandElement::DECISION;
	
	//separate decision and consequnces
	string curElement;
	int depth = 0;
	bool foundQuestion = false;
	bool foundColon = false;
	char prevChar = ' ';
	for(size_t pos = 0; pos < cmd.length(); ++pos)
	{
		char curChar = cmd[pos];
		if(IsControlChar(curChar, prevChar))
		{
			if(depth == 0)
			{
				switch(curChar)
				{
				case '?':
					if(foundQuestion)
						throw CF_ParseException("CF_CEDecision::CF_CEDecision()", "Syntax error in command");
					decision = CreateCommandElement(curElement);
					curElement.erase();
					foundQuestion = true;
					break;

				case ':':
					if(!foundQuestion || foundColon)
						throw CF_ParseException("CF_CEDecision::CF_CEDecision()", "Syntax error in command");
					consequence[0] = CreateCommandElement(curElement);
					curElement.erase();
					foundColon = true;
					break;

				case '{':
					++depth;
					curElement+= curChar;
					break;

				default:
					throw CF_ParseException("CF_CEDecision::CF_CEDecision()", "Syntax error in command");
				}//switch(curChar)
			}//if(depth == 0)
			else
			{
				switch(curChar)
				{
				case '{':
					++depth;
					break;
					
				case '}':
					--depth;
					break;
					
				default:
					break;
				}//switch(curChar)
				curElement += curChar;
			}//else
		}//if(IsControlChar(curChar, prevChar))
		else
		{
			curElement += curChar;
		}//else
		prevChar = curChar;
	}//for(size_t pos = 0; pos < cmd.length(); ++pos)
	if(!foundColon || !foundQuestion)
		throw CF_ParseException("CF_CEDecision::CF_CEDecision()", "Syntax error in command");
		
	//whatever is left in curElement is the false consequence
	consequence[1] = CreateCommandElement(curElement);
}//CF_CEDecision::CF_CEDecision(string cmd)

//CF_CEDecision::~CF_CEDecision: Gets rid of created command elements
//returns: <nothing>
CF_CEDecision::~CF_CEDecision()
{
	if(decision != NULL)
		delete decision;
	if(consequence[0] != NULL)
		delete consequence[0];
	if(consequence[1] != NULL)
		delete consequence[1];
}//CF_CEDecision::~CF_CEDecision()

//CF_CEDecision::Expand: evaluates the decision and expands the approprties consequence
//returns: epxanded consequence
string CF_CEDecision::Expand()
{
	string decStr = decision->Expand();
	if(decStr.empty())
	{
		return consequence[1]->Expand();
	}//if((decision.Expand()).empty())
	
	return consequence[0]->Expand();
}//string CF_CEDecision::Expand()

void CF_CEDecision::Execute()
{
	string decStr = decision->Expand();
	if(decStr.empty())
	{
		return consequence[1]->Execute();
	}//if((decision.Expand()).empty())
	
	return consequence[0]->Execute();
}//void CF_CEDecision::Execute()


//END CF_CEDecision


/////////////////////////////////////////////////////
//BEGIN CF_CEList

//CF_CEList::CF_CEList: Creates a list command element
//	cmdStr: string to parse
//returns: <nothing>
CF_CEList::CF_CEList(const list<pair<string, Type> >& eList)
{
	//set type
	type = CF_CommandElement::LIST;
	
	//create elements in list
	for(list<pair<string, Type> >::const_iterator it = eList.begin(); it != eList.end(); ++it)
	{
		CF_CommandElement* newElement = NULL;
		switch((*it).second)
		{
		case STRING:
			newElement = reinterpret_cast<CF_CommandElement*>(new CF_CEString((*it).first));
			break;
			
		case SHELL:
			newElement = reinterpret_cast<CF_CommandElement*>(new CF_CEShell((*it).first));
			break;
			
		case BUILTIN:
			newElement = reinterpret_cast<CF_CommandElement*>(new CF_CEBuiltIn((*it).first));
			break;
			
		case VARIABLE:
			newElement = reinterpret_cast<CF_CommandElement*>(new CF_CEVariable((*it).first));
			break;
			
		case DECISION:
			newElement = reinterpret_cast<CF_CommandElement*>(new CF_CEDecision((*it).first));
			break;			
		
		case LIST:
			throw CF_ParseException("CF_CommandElement::CreateCommandElement()", "Syntax error in command");
			break;
		}//switch((*it).second)
		
		CF_Assert(newElement != NULL, "CF_CEList::CF_CEList()");
		elements.push_back(newElement);
	}//for(list<pair<string, Type> >::iterator it = eList.begin(); it != eList.end(); ++it)
}//CF_CEList::CF_CEList(string cmdStr)

//CF_CEList::~CF_CEList: deletes element list
//returns: <nothing>
CF_CEList::~CF_CEList()
{
	while(!elements.empty())
	{
		if(elements.front() != NULL)
			delete elements.front();
		elements.pop_front();
	}//while(!elements.empty())
}//CF_CEList::~CF_CEList()

//CF_CEList::Expand: expands command elements in list
//returns: expanded string
string CF_CEList::Expand()
{
	string output;
	for(list<CF_CommandElement*>::iterator it = elements.begin(); it != elements.end(); ++it)
	{
		output += (*it)->Expand();
	}//for(elements.iterator it = elements.begin(); it != elements.end(); ++it)
	
	return output;
}//string CF_CEList::Expand()


//END CF_CEList


/////////////////////////////////////////////////////
//BEGIN CF_CEBuiltIn

//CF_CEBuiltIn::CF_CEBuiltIn: creates a built in command element
//	cmd: string to parse
//returns: <nothing>
CF_CEBuiltIn::CF_CEBuiltIn(const string& cmd)
{
	//set type
	type = CF_CommandElement::BUILTIN;
	
	//save cmd mgr
	cmdMgr = curCmdMgr;
	
	//find command name
	int depth = 0;
	bool foundName = false;
	string curElement;
	char prevChar = ' ';
	for(size_t pos = 0; pos < cmd.length(); ++pos)
	{
		char curChar = cmd[pos];
		if(!IsControlChar(curChar, prevChar))
		{
			curElement += curChar;
		}//if(prevChar == '\\')
		else
		{
			if(depth == 0)
			{
				switch(curChar)
				{
				case '(':
					if(foundName)
						throw CF_ParseException("CF_CEBuiltIn::CF_CEBuiltIn()", "Syntax error in command");
					function = curElement;
					curElement.erase();
					foundName = true;
					break;
				
				case ',':
					if(!foundName)
						throw CF_ParseException("CF_CEBuiltIn::CF_CEBuiltIn()", "Syntax error in command");
					args.push_back(CreateCommandElement(curElement));
					curElement.erase();
					break;

				case ')':
					if(!foundName || (pos != (cmd.length() - 1)))
						throw CF_ParseException("CF_CEBuiltIn::CF_CEBuiltIn()", "Syntax error in command");
					args.push_back(CreateCommandElement(curElement));
					curElement.erase();
					break;
					
				case '{':
					if(!foundName)
						throw CF_ParseException("CF_CEBuiltIn::CF_CEBuiltIn()", "Syntax error in command");
					++depth;
					curElement += curChar;
					break;
				
				case '}':
				case '!':
				case ':':
				case '?':
				default:
					throw CF_ParseException("CF_CEBuiltIn::CF_CEBuiltIn()", "Syntax error in command");
				}//switch(curChar)				
			}//if(depth == 0)
			else
			{
				switch(curChar)
				{
				case '{':
					++depth;
					break;
					
				case '}':
					--depth;
					break;
					
				default:
					break;
				}//switch(curChar)
				curElement += curChar;
			}//else
		}//else
		prevChar = curChar;
	}//for(size_t pos = 0; pos < cmd.length(); ++pos)
}//CF_CEBuiltIn::CF_CEBuiltIn(string cmd)

//CF_CEBuiltIn::~CF_CEBuiltIn: gets rid of created command elements
//returns: <nothing>
CF_CEBuiltIn::~CF_CEBuiltIn()
{
	while(!args.empty())
	{
		if(args.back() != NULL)
			delete args.back();
		args.pop_back();
	}//while(!args.empty())
}//CF_CEBuiltIn::~CF_CEBuiltIn()

//CF_CEBuiltIn::Expand: Calls function with expanded args
//returns: output from function
string CF_CEBuiltIn::Expand()
{
	//expand args
	vector<string> strArgs;
	for(vector<CF_CommandElement*>::iterator it = args.begin(); it != args.end(); ++it)
	{
		strArgs.push_back((*it)->Expand());
	}//for(agrs.iterator it = args.begin(); it != args.end(); ++it)
	
	//get function
	return cmdMgr->CallFunction(function, strArgs);
}//string CF_CEBuiltIn::Expand()


//END CF_CEBuiltIn


/////////////////////////////////////////////////////
//BEGIN CF_CEShell

//CF_CEShell::CF_CEShell: Creeates a shell command element
//	cmd: string to parse
//returns: <nothing>
CF_CEShell::CF_CEShell(const string& cmd)
{
	//set type
	type = CF_CommandElement::SHELL;

	//create shell string command element
	cmdElement = CreateCommandElement(cmd);
}//CF_CEShell::CF_CEShell(string cmd)

//CF_CEShell::~CF_CEShell: gets rid of created cmd element
//returns: <nothing>
CF_CEShell::~CF_CEShell()
{
	if(cmdElement != NULL)
		delete cmdElement;
}//CF_CEShell::~CF_CEShell()

//CF_CEShell::Expand: runs shell command 
//returns: output from shell command
string CF_CEShell::Expand()
{
	//expand shell cmd
	string shellCmd = cmdElement->Expand();
	
	//open pipe
	FILE* pipe = popen(shellCmd.c_str(), "r");
	if(pipe == NULL)
	{
		cerr << "Couldn't open pipe to command:" << shellCmd << endl;
		return "";
	}//if(pipe == NULL)
	
	//read output
	char buffer[256];
	string output;
	while(fgets(buffer, sizeof(buffer), pipe))
		output += buffer;
		
	//close pipe
	int res = pclose(pipe);
	if(res == 32512)
	{
		cerr << "Shell error executing " << shellCmd << endl;
	}//if(res == 32512)
	else if(res != 0)
		throw CF_Exception("CF_CEShell:Expand()", "Error closing pipe");
	
	return output; 
}//string CF_CEShell::Expand()

static list<pid_t> pids;
static bool watchingPids;
gint PidWatch(gpointer)
{
	if(!pids.empty())
	{
		list<pid_t> dead;
		for(list<pid_t>::iterator it = pids.begin(); it != pids.end(); ++it)
		{
			int status;
			pid_t pid = waitpid(*it, &status, WNOHANG);
			if(pid != 0)
			{
				dead.push_back(pid);
			}//if(pid != 0)
		}//for(list<int>::iterator it = pids.begin(); it != pids.end(); ++it)

		for(list<pid_t>::iterator it = dead.begin(); it != dead.end(); ++it)
		{
			pids.remove(*it);
		}//for(list<int>::iterator it = pids.begin(); it != pids.end(); ++it)
		
		return TRUE;
	}//if(!pids.empty())
	else 
	{
		watchingPids = false;
		return FALSE; 
	}//else 
}//static gint PidWatch()


void CF_CEShell::Execute()
{
	//expand cmd first (in case cmd uses GUI to get info)	
	string shellCmd = cmdElement->Expand();
	
	//fork process
	pid_t pid = fork();
	
	//if parent start watching for child to exit
	if(pid == -1)
	{
		cerr << "Couldn't fork process" << endl;
		return;
	}//if(pid != 0)
	else if(pid != 0)
	{
		pids.push_back(pid);
		if(!watchingPids)
		{
			watchingPids = true;
			gtk_timeout_add(500, PidWatch, NULL);
		}//if(!watchingPids)	
	}//else if(pid != 0)
	else
	{
		//exec command
		execl("/bin/sh", "/bin/sh", "-c", shellCmd.c_str(), NULL);

		cerr << "Failure Executing:" << shellCmd << ":" << endl;
		cerr << "error: " << strerror(errno) << endl;
		//if we get here something went wrong
		throw CF_Exception("CF_CEShell::Execute()", "Couldn't exec shell command");
	}//else
}//void CF_CEShell::Execute()


//END CF_CEShell




//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//BEGIN CEList

void CEList::AddCommand(CF_CommandElement* cmd)
{
	ceList.push_back(cmd);
}//void CF_ToolbarWidget::CEList::AddCommand(CF_CommandElement* cmd)

void CEList::ExpandCommands()
{
	for(list<CF_CommandElement*>::iterator it = ceList.begin(); it != ceList.end(); ++it)
		(*it)->Execute();	
}//void CF_ToolbarWidget::CEList::ExpandCommands()

CEList::~CEList()
{
	while(!ceList.empty())
	{
		if(ceList.front() != NULL)
			delete ceList.front();
		ceList.pop_front();
	}//while(!ceList.empty())	
}//CF_ToolbarWidget::CEList::~CEList()

//END CEList

//////////////////////////////////////////////////
//BEGIN GLOBALS

//IsFunction: Checks whether string is a function or not
//	cmdStr: string to check 
//returns: true if stringg is a function
static bool IsFunction(const string& cmdStr)
{
	//parse string for "()" outside of "{}"
	int depth = 0;
	bool fnc = false;
	char prevChar = ' ';
	for(unsigned int pos = 0; pos < cmdStr.length(); ++pos)
	{
		char curChar = cmdStr[pos];
		if((curChar == '(') && (prevChar != '\\') && (depth == 0))
		{
			fnc = true;
			break;
		}//if((curChar == '(') && (depth == 0))
		if((curChar == '{')	&& (prevChar != '\\'))
			++depth;
		else if((curChar == '}') && (prevChar != '\\'))
			--depth;
		prevChar = curChar;
	}//for(int pos = 0; pos < cdmStr.length(); ++pos)
	
	return fnc;
}//static bool IsFunction(const string& cmdStr)

//END GLOBALS


