/*****************************************************************************/
//  cf_command.hh
//  Copyright (C) 1999 Brett Hall <swizin@rain.org>
//  
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or 
//  (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 USA 
/*****************************************************************************/


//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// cf_command.hh
//	-Command and command elemtent  classes
//	-Command classes implement commands
//	-Command element objects are combined by the commands to create full commands
//		-command elements implement expansion of commands and variables
//
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

#ifndef CF_COMMAND_HH
#define CF_COMMAND_HH

#include <list>
#include <string>
#include <vector>
#include "cf_callback.hh"
#include "cf_scriptvar.hh"

using namespace std;

class CF_CommandElement;	//forward declaration needed by CEList

//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//CEList - list of command elements

class CEList
{
public:
	void AddCommand(CF_CommandElement* cmd);
	void ExpandCommands();
	~CEList();

private:
	list<CF_CommandElement*> ceList;
};//class CEList


//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//CF_CommandElement - abstract base class for command elements

/**The base class for all command elements.
	Commands from scirpt files are built from command elementss by calling their Expand() method
	and concatenating all the strings that are returned.  Command elemtns are created by calling
	the static method CreateCommandElement() with the command string to be processed.  The returned
	element contains all the other elements making up the command, calling expand on this element
	will expand all the others.
*/
class CF_CommandElement
{
public:
	/**Enumerates type of element.
		possible values:
		<ll>
		<li><b>STRING</b> - a string element
		<li><b>VARIABLE</b> - a script var element
		<li><b>BUILTIN</b> - a "built-in" command
		<li><b>SHELL</b> - a shell command
		<li><b>DECISION</b> - a decision(ie <blah ? res1 : res2>)
		<li><b>LIST</b> - a list of other elements
		</ll>
	*/
	enum Type {STRING, VARIABLE, BUILTIN, SHELL, DECISION, LIST};
	
	//Virtual destructor does nothing
	virtual ~CF_CommandElement()	{}
	
	/**"Expands" the command element, carries out whatever operation the element represents
		@return The result of the command carries out
	*/
	virtual string Expand() = 0;
	
	/**Executes the command element but doesn't wait for output.  Used for calling commands from 
		top level of script, won't hang on shell programs that detach from control freak
	*/
	virtual void Execute() = 0;
	
	/**Gets the type of the element
		@return The element type
	*/
	Type GetType()	{return type;}
	
	/**Creates a command element (and all elements making it up)
		@param cmdStr String to parse when creating command
		@param cmdMgr "Built-in" command manager to use for elements that need it
		@param varMgr Script var manager to use for elements that need it
		@throws CF_ParseException if sytax error is found in script 
	*/
	static CF_CommandElement* CreateCommandElement(const string& cmdStr, CF_CommandMgr* cmdMgr, CF_VarMgr* varMgr);
	
	/**Creates a command elemnts in string
		@param cmdStr String to parse when creating command
		@param cmds List to put commands in
		@param functions Built in command manager to use
		@param varMgr Script var manager to use for elements that need it
	*/
	static void CreateCommands(const string& cmdStr, CEList& cmds, CF_CommandMgr* functions, CF_VarMgr* varMgr);
	
protected:
	static CF_CommandElement* CreateCommandElement(const string& cmdStr);
	
	Type type;
	static CF_VarMgr* curVarMgr;
	static CF_CommandMgr* curCmdMgr;
};

//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// class CF_CEList

///A list of other command elements
class CF_CEList : public CF_CommandElement
{
public:
	///creates a command element list
	CF_CEList(const list<pair<string, Type> >& eList);
	
	//detroys the elements in the list
	~CF_CEList();
	
	/**Expands all elements in list
		@return Result of expansion of elements in list
	*/
	string Expand();
	
	void Execute() {}

private:
	list<CF_CommandElement*> elements;
};//class CF_CEList

//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// class CF_CEString

///Element that expands to a static string
class CF_CEString : public CF_CommandElement
{
public:
	/**Creates string element
		@param cmdStr String that element will expand to
	*/
	CF_CEString(const string& cmdStr);
	
	///Returns the string contained in element
	string Expand() {return str;}
	
	void Execute() {}

private:
	string str;
};//class CF_CEString

//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// class CF_CEVariable

///Expands to constents of a script var 
class CF_CEVariable : public CF_CommandElement
{
public:
	/**Creates a var element
		@param cmdStr Name of variable to expand
	*/
	CF_CEVariable(const string& cmdStr);
	
	//destroys var name element
	~CF_CEVariable();
	
	///Returns contents of var
	string Expand();
	
	void Execute() {}
	
private:
	CF_CommandElement* varName;
	CF_VarMgr* varMgr;
};//class CF_CEVariable


//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// class CF_CEDecision

/**A <i>decision</i> element.
	<i>Decision</i> elements are of the form "dec ? res1 : res2", dec is evaluated, if true res1
	is expanded, if false res2 is.  dec, res1 and res2 are all command elements as well
*/
class CF_CEDecision : public CF_CommandElement
{
public:
	/**Creates a decision element
		@param cmdStr The decision string to parse
	*/
	CF_CEDecision(const string& cmd);
	
	//Destroys command elements for decision and consequences
	~CF_CEDecision();
	
	///Expands decision, if true expands and returns consequnce 1 else consequence 2
	string Expand();
	
	void Execute();
	
private:
	CF_CommandElement* consequence[2];
	CF_CommandElement* decision;
};//class CF_CEVarDecision


//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// class CF_ShellCommand

///Expands to output of a shell command
class CF_CEShell : public CF_CommandElement
{
public:
	/**Creates a shell command element
		@param cmdStr Shell command to use
	*/
	CF_CEShell(const string& cmd);
	
	//Destroys shell string command element
	~CF_CEShell();
	
	/**Expands and executes shell command element
		@return Output from shell command
	*/
	string Expand();
	
 	void Execute();

private:
	CF_CommandElement* cmdElement;
};//class CF_ShellCommand

//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// class CF_BuiltInCommand

///Expands to the output from a "built-in" command
class CF_CEBuiltIn : public CF_CommandElement
{
public:
	/**Creates a <i>built-in</i> command element
		@param cmd String to parse for built in command
	*/
	CF_CEBuiltIn(const string& cmd);
	
	//Destroys command and args elements
	~CF_CEBuiltIn();
	
	/**Executes <i>built-in</i> command using expanded arguments
		@return Output from command
	*/
	string Expand();
	
	void Execute() {Expand();}

private:
	vector<CF_CommandElement*> args;
	string function;
	CF_CommandMgr* cmdMgr;
};//class CF_BuiltInCommand




#endif
