/*****************************************************************************/
//  cf_toolbarwidget.cc
//  Copyright (C) 1999 Brett Hall <swizin@rain.org>
//  
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or 
//  (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 USA 
/*****************************************************************************/


//////////////////////////////////////////////////
// cf_toolbarwidget.cc
//
//////////////////////////////////////////////////


#include "cf_toolbarwidget.hh"
#include "cf_exception.hh"
#include "pathinfo.hh"
#include <stack>

//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//Constants

//shared name tokens
static const size_t nameToken = 0;

//button tokens
static const size_t numButtonTokens = 4;
static const size_t buttonIconToken = 1;
static const size_t buttonToolTipToken = 2;
static const size_t buttonCmdToken = 3;

//toggle tokens
static const size_t numToggleTokens = 6;
static const size_t toggleIconToken = 1;
static const size_t toggleToolTipToken = 2;
static const size_t toggleVarToken = 3;
static const size_t toggleDownCmdToken = 4;
static const size_t toggleUpCmdToken = 5;

//text tokens
static const size_t numTextTokens = 6;
static const size_t textToolTipToken = 1;
static const size_t textWidthToken = 2;
static const size_t textVarToken = 3;
static const size_t textInitToken = 4;
static const size_t textCmdToken = 5;

//combo tokens
static const size_t numComboTokens = 8;
static const size_t comboToolTipToken = 1;
static const size_t comboWidthToken = 2;
static const size_t comboVarToken = 3;
static const size_t comboInitBoxToken = 4;
static const size_t comboInitListToken = 5;
static const size_t comboReadOnlyToken = 6;
static const size_t comboCmdToken = 7;


//////////////////////////////////////////////////
//BEGIN CF_ToolbarWidget

CF_CommandMgr* CF_ToolbarWidget::functions = NULL;
CF_VarMgr* CF_ToolbarWidget::varMgr = NULL;
list<string> CF_ToolbarWidget::iconPaths;

void CF_ToolbarWidget::SetIconPath(string path)
{
	//split path string into separate paths
	bool done = false;
	while(!done)
	{
		size_t pos = path.find(':');
		if(pos == string::npos)
		{
			iconPaths.push_back(path);
			break;
		}//if(pos == string::npos)
		string newPath = path.substr(0, pos);
		path = path.substr(pos, string::npos);
		iconPaths.push_back(newPath);
		if(path.empty())
			done = true;
		else
			path.erase(0, 1);
	}//while(!done)
}//void CF_ToolbarWidget::SetIconPath(const string& path)


//CF_ToolbarWidget::TokenizeInitStr: splits inialization string into tokens
//	initStr: initializer string from script file
//	tokens: list of tokens from initStr
//returns: <nothing>
void CF_ToolbarWidget::TokenizeInitStr(const string& initStr, vector<string>& tokens)
{
	//make sure all {} and () are matched
	VerifyMatching(initStr);
	
	//split into tokens
	string token;
	int depth = 0;
	for(size_t pos = 0; pos < initStr.length(); ++pos)
	{
		switch(initStr[pos])
		{
		case ',':
			if(depth == 0)
			{
				if(token.empty())
					throw CF_ParseException("CF_ToolbarWidget::TokenizeInitStr()", "Empty token error");
				tokens.push_back(token);
				token = "";
			}//if(depth == 0)
			else
			{
				token += initStr[pos];
			}//else
			break;
			
		case '{':
		case '(':
			++depth;
			token += initStr[pos];
			break;
			
		case '}':
		case ')':
			--depth;
			token += initStr[pos];
			break;
			
		default:
			token += initStr[pos];
			break;
		}//switch(initStr[pos])
	}//for(int pos = 0; pos < initStr.length(); ++pos)
	if(!token.empty())
		tokens.push_back(token);
}//void CF_ToolbarWidget::TokenizeInitStr(const string& initStr, list<string>& tokens)


//CF_ToolbarWidget::VerifyMatching: verifies that all {} and () are matched
//	str: string to parse
//returns: <nothing> (throws parse exception if things aren't OK)
void CF_ToolbarWidget::VerifyMatching(const string& str)
{
	//make sure all {} and () are matched
	stack<int> depth;	
	depth.push(0);
	for(size_t pos = 0; pos < str.length(); ++pos)
	{
		char cur = str[pos];
		switch(cur)
		{
		case '{':
			++(depth.top());
			break;
			
		case '}':
			--(depth.top());
			break;
			
		case '(':
			depth.push(0);
			break;
			
		case ')':
			if(depth.top() != 0)
				throw CF_ParseException("CF_ToolbarWidget::VerifyMatching()", "Mismatched {}");
			depth.pop();
			if(depth.empty())
				throw CF_ParseException("CF_ToolbarWidget::VerifyMatching()", "Mismatched ()");
			break;
			
		default:	
			break;
		}//switch(cur)
	}//for(int pos = 0; pos < initStr.length(); ++pos)
	if(depth.top() != 0)
		throw CF_ParseException("CF_ToolbarWidget::VerifyMatching()", "Mismatched {}");
	if(depth.size() != 1)
		throw CF_ParseException("CF_ToolbarWidget::VerifyMatching()", "Mismatched ()");	
}//bool CF_ToolbarWidget::VerifyMatching(const string& str)


string CF_ToolbarWidget::FindIcon(const string& iconName)
{
	//try current working directory
	ifstream fileTest(iconName.c_str(), ios::nocreate);
	if(!fileTest){/*do nothing*/}
	else
	{
		//file exists
		fileTest.close();
		return iconName;
	}//else
	
	//search icon paths first
	for(list<string>::iterator it = iconPaths.begin(); it != iconPaths.end(); ++it)
	{
		//create icon path-name
		string file = *it;
		if(file[file.length() - 1] != '/')
			file += "/";
		file += iconName;
		
		//check for file existence
		ifstream fileTest(file.c_str(), ios::nocreate);
		if(!fileTest){/*do nothing*/}
		else
		{
			//file exists
			fileTest.close();
			return file;
		}//else
	}//for(list<string>::iterator it = iconsPaths.begin(); it != iconPAths.end(); ++it)
	
	//search other paths
	string file = PathInfo::GetInstance().BuildPathName(iconName);
	if(!file.empty())
		return file;
	
	//couldn't find icon
	return "";
}//string CF_ToolbarWidget::FindIcon(const strring& iconName)

//END CF_ToolbarWidget



/////////////////////////////////////////////////////
//BEGIN CF_Button

//CF_Button::~CF_Button: 
//returns: <nothing>
CF_Button::~CF_Button()
{
	if(button != NULL)
		delete button;
	if(icon != NULL)
		delete icon;
}//CF_Button::~CF_Button()


//CF_Button::Clicked: connected to button's clicked signal
//returns: <nothing>
void CF_Button::Clicked()
{
	cmds.ExpandCommands();
}//void CF_Button::Clicked()

//CF_Button::CF_Button: creates button based on line in script file
//	initStr: initailizatoin string from script file
//	label: use a label if true
//	icon: use am icon if true
//returns: <nothing>
CF_Button::CF_Button(const string& initStr, Gtk_Toolbar& bar, Gtk_Window& win)
{
	//init members
	button = NULL;
	icon = NULL;
	
	//tokenize init string
	vector<string> tokens;
	TokenizeInitStr(initStr, tokens);
	if(tokens.size() != numButtonTokens)
		throw CF_ParseException("CF_Button::CF_Button()", "Wrong number of parameters for button");
		
	//set type and name
	type = BUTTON;
	name = tokens[nameToken];
	
	//create cmd list
	CF_CommandElement::CreateCommands(tokens[buttonCmdToken], cmds, functions, varMgr);
	
	//create icon
	bool iconFound = false;
	string iconFile = FindIcon(tokens[buttonIconToken]);
	if(!iconFile.empty())
		iconFound = true;
	
	//create the button
	if(iconFound)
	{
		button = new Gtk_Button;
		CF_Assert(button != NULL, "CF_Button::CF_Button()");
		//button->show();
		icon = new Gtk_Pixmap(win, iconFile);
		CF_Assert(icon != NULL, "CF_Button::CF_Button()");
		icon->show();
		button->add(*icon);
		button->show();
	}//if(foundIcon)
	else
	{
		button = new Gtk_Button(tokens[nameToken]);
		CF_Assert(button != NULL, "CF_Button::CF_Button()");
		button->show();
		cerr << "ControlFreak: Couldn't find icon " << tokens[buttonIconToken] << endl;	
	}//else
	
	//set up click handler
	connect_to_method(button->clicked, this, &CF_Button::Clicked);
	
	//show and append button
	bar.append_widget(*button, tokens[buttonToolTipToken]);
}//CF_Button::CF_Button(const string& initLine)


//END CF_Button


/////////////////////////////////////////////////////
//BEGIN CF_Toggle

CF_Toggle::CF_Toggle(const string& initStr, Gtk_Toolbar& bar, Gtk_Window& win)
{
	//init members
	button = NULL;
	icon = NULL;
	disableCmd = false;
	
	//tokenize init string
	vector<string> tokens;
	TokenizeInitStr(initStr, tokens);
	if(tokens.size() != numToggleTokens)
		throw CF_ParseException("CF_Button::CF_Button()", "Wrong number of parameters for toggle");
		
	//set type and name
	type = TOGGLE;
	name = tokens[nameToken];
	
	//create cmd list
	if(tokens[toggleUpCmdToken] != "N/A")
		CF_CommandElement::CreateCommands(tokens[toggleUpCmdToken], upCmd, functions, varMgr);
	if(tokens[toggleDownCmdToken] != "N/A")
		CF_CommandElement::CreateCommands(tokens[toggleDownCmdToken], downCmd, functions, varMgr);
	
	//get state var
	stateVar = varMgr->GetVar(tokens[toggleVarToken]);
	if(stateVar == NULL)
		throw CF_ParseException("CF_Toggle::CF_toggle()", string("Using non-existent var ") + tokens[toggleVarToken]);
	stateVar->SetValue(false);
	
	//create icon
	bool iconFound = false;
	string iconFile = FindIcon(tokens[toggleIconToken]);
	if(!iconFile.empty())
		iconFound = true;
	
	//create the button
	if(iconFound)
	{
		button = new Gtk_ToggleButton;
		CF_Assert(button != NULL, "CF_Button::CF_Button()");
		button->show();
		Gdk_Bitmap mask;
		Gdk_Pixmap xpm(win.get_window(), mask,
						 Gdk_Color(&win.get_style()->gtkobj()->bg[GTK_STATE_NORMAL]), iconFile);
		icon = new Gtk_Pixmap(xpm, mask);
		CF_Assert(icon != NULL, "CF_Button::CF_Button()");
		icon->show();
		button->add(*icon);
	}//if(foundIcon)
	else
	{
		button = new Gtk_ToggleButton(tokens[nameToken]);
		CF_Assert(button != NULL, "CF_Button::CF_Button()");
		button->show();
		cerr << "ControlFreak: Couldn't find icon " << tokens[buttonIconToken] << endl;	
	}//else
	
	//set up click handler
	connect_to_method(button->toggled, this, &CF_Toggle::Toggle);
	
	//show and append button
	bar.append_widget(*button, tokens[toggleToolTipToken]);	
}//CF_Toggle::CF_Toggle(const string& initStr, Gtk_Toolbar& bar, Gtk_Window& win)

CF_Toggle::~CF_Toggle()
{
	if(icon != NULL)
		delete icon;
		
	if(button != NULL)
		delete button;
}//CF_Toggle::~CF_Toggle()

void CF_Toggle::Toggle()
{
	if(button->get_active())
	{
		stateVar->SetValue(true);
		if(!disableCmd)
			downCmd.ExpandCommands();
		else
			disableCmd = false;
	}//if(button->get_active())
	else
	{
		stateVar->SetValue(false);
		if(!disableCmd)
			upCmd.ExpandCommands();
		else
			disableCmd = false;
	}//else
}//void CF_Toggle::Toggle()

void CF_Toggle::SetState(const string& state)
{
	if(state.empty() || (state == "false"))
	{
		//set toggle to false
		button->set_active(false);
	}//if(state.empty())
	else
	{
		//set to true
		button->set_active(true);
	}//else
}//void CF_Toggle::SetState(const string& state)


void CF_Toggle::WriteState(ofstream& file)
{
	string output = button->get_active() ? "true" : "false";
	file << output << endl;
}//void CF_Toggle::WriteState(ofstream& file)

void CF_Toggle::ReadState(ifstream& file)
{
	string input;
	getline(file, input);
	disableCmd = true;
	button->set_active(input == "true");
}//void CF_Toggle::ReadState(ifstream& file)


//END CF_Toggle


/////////////////////////////////////////////////////
//BEGIN CF_Text

CF_Text::CF_Text(const string& initStr, Gtk_Toolbar& bar)
{
	//init members
	entry = NULL;
	
	//tokenize init string
	vector<string> tokens;
	TokenizeInitStr(initStr, tokens);
	if(tokens.size() != numTextTokens)
		throw CF_ParseException("CF_Text::CF_Text()", "Wrong number of parameters for text entry");
		
	//set type and name
	type = TEXT;
	name = tokens[nameToken];
	
	//create cmd list
	if(tokens[textCmdToken] != "N/A")
		CF_CommandElement::CreateCommands(tokens[textCmdToken], cmds, functions, varMgr);
		
	//get state var
	stateVar = varMgr->GetVar(tokens[textVarToken]);
	if(stateVar == NULL)
		throw CF_ParseException("CF_Text::CF_Text()", string("Using non-existent var ") + tokens[textVarToken]); 
		
	//create the entry
	entry = new Gtk_Entry;
	CF_Assert(entry != NULL, "CF_Text::CF_text()");
	if(tokens[textInitToken] != "N/A")
	{
		entry->set_text(tokens[textInitToken]);
		stateVar->SetValue(tokens[textInitToken]);	
	}//if(tokens[textInitToken] != "NONE")
	else
	{
		entry->set_text("");
		stateVar->SetValue(string(""));		
	}//else
	if(tokens[textWidthToken] != "N/A")
	{
		int width = atoi(tokens[textWidthToken].c_str());
		entry->set_usize(width, -1);
	}//if(tokens[textWidthToken] != "N/A")
	entry->show();
	
	//set up click handler
	connect_to_method(entry->changed, this, &CF_Text::Changed);
	connect_to_method(entry->activate, this, &CF_Text::Activate);
	
	//show and append button
	bar.append_widget(*entry, tokens[textToolTipToken]);	
}//CF_Text::CF_Text(const string& initStr, Gtk_Toolbar& bar, Gtk_Window& win)

CF_Text::~CF_Text()
{
	if(entry != NULL)
		delete entry;
}//CF_Text::~CF_Text()

void CF_Text::Changed()
{
	stateVar->SetValue(entry->get_text());
}//void CF_Text::Change()

void CF_Text::Activate()
{
	cmds.ExpandCommands();
}//void CF_Text::Actviate()

void CF_Text::WriteState(ofstream& file)
{
	file << entry->get_text() << endl;
}//void CF_Text::WriteState(ofstream& file)

void CF_Text::ReadState(ifstream& file)
{
	string state;
	getline(file, state);
	entry->set_text(state);
}//void CF_Text::ReadState(ifstream& file)

//END CF_Text


/////////////////////////////////////////////////////
//BEGIN CF_Combo

CF_Combo::CF_Combo(const string& initStr, Gtk_Toolbar& bar)
{
	//init members
	combo = NULL;
	tip = NULL;
	
	//tokenize init string
	vector<string> tokens;
	TokenizeInitStr(initStr, tokens);
	if(tokens.size() != numComboTokens)
		throw CF_ParseException("CF_Combo::CF_Combo()", "Wrong number of parameters for combo box");
		
	//set type and name
	type = COMBO;
	name = tokens[nameToken];
	
	//create cmd list
	if(tokens[comboCmdToken] != "N/A")
		CF_CommandElement::CreateCommands(tokens[comboCmdToken], cmds, functions, varMgr);
		
	//get state var
	stateVar = varMgr->GetVar(tokens[comboVarToken]);
	if(stateVar == NULL)
		throw CF_ParseException("CF_Combo::CF_Combo()", string("Using non-existent var ") + tokens[comboVarToken]); 
		
	//create the entry
	combo = new Gtk_Combo;
	CF_Assert(combo != NULL, "CF_Combo::CF_Combo()");
	combo->disable_activate();
	if(tokens[comboInitBoxToken] != "N/A")
		stateVar->SetValue(tokens[comboInitBoxToken]);
	else
		stateVar->SetValue(string(""));
	if(tokens[comboReadOnlyToken] == "true")
		combo->get_entry()->set_sensitive(false);
	if(tokens[comboWidthToken] != "N/A")
	{
		int width = atoi(tokens[comboWidthToken].c_str());
		combo->set_usize(width, -1);
	}//if(tokens[comboWidthToken] != "N/A")
	combo->show();
	
	//fill the list
	string init = tokens[comboInitListToken];
	if(init != "N/A")
	{
		if(init[0] == '{')
		{
			if(init[init.length()- 1] != '}')
				throw CF_ParseException("CF_Combo::CF_Combo()", "Can't mix expansion in combo list initializer");

			init.erase(0, 1);
			init.erase(init.length()- 1, 1);
			CF_ScriptVar* var = varMgr->GetVar(init);
			if(var == NULL)
				throw CF_ParseException("CF_Combo::CF_Combo()", "Bad variable name in list initializer");
			init = varMgr->GetVar(init)->GetStringValue();
		}//if(tokens[comboInitListToken][0] == '{')
		string curStr;
		for(size_t pos = 0; pos < init.length(); ++pos)
		{
			switch(init[pos])
			{
			case '|':
				comboStrings.push_back(curStr);
				curStr = "";
				break;
			default:
				curStr += init[pos];
				break;
			}//switch(init[pos])
		}//for(size_t pos = 0; pos < init.length(); ++pos)
		if(!curStr.empty())
			comboStrings.push_back(curStr);
		if(tokens[comboInitBoxToken] != "N/A")
			SetEntry(tokens[comboInitBoxToken], true);	
		else
		{
			GList* cList = NULL;
			for(list<string>::iterator it = comboStrings.begin(); it != comboStrings.end(); ++it)
				cList = g_list_append(cList, const_cast<char*>((*it).c_str()));
			combo->set_popdown_strings(cList);
			g_list_free(cList);
			combo->get_entry()->set_text("");
		}//else
	}//if(init != "NONE")
	
	//set up click handler
	connect_to_method(combo->get_entry()->changed, this, &CF_Combo::Changed);
	connect_to_method(combo->get_entry()->activate, this, &CF_Combo::Activate);
	
	//create tool tip 
	if(!tokens[comboToolTipToken].empty())
	{
		tip = new Gtk_Tooltips;
		CF_Assert(tip != NULL, "CF_Combo::CF_Combo()");
		tip->set_tip(*(combo->get_entry()), tokens[comboToolTipToken].c_str());
		tip->set_tip(*(combo->get_button()), tokens[comboToolTipToken].c_str());
	}//if(!tokens[comboToolTipToken].empty())
	
	//show and append button
	bar.append_widget(*combo);		
}//CF_Combo::CF_Combo(const string& initStr, Gtk_Toolbar& bar, Gtk_Window& win)

CF_Combo::~CF_Combo()
{
	if(combo != NULL)
		delete combo;
	if(tip != NULL)
		delete tip;
}//CF_Combo::~CF_Combo()

void CF_Combo::Changed()
{
	stateVar->SetValue(combo->get_entry()->get_text());
}//void CF_Combo::Changed()

void CF_Combo::Activate()
{
	SetEntry(stateVar->GetStringValue(), false);
	cmds.ExpandCommands();
}//void CF_Combo::Activate()

void CF_Combo::WriteState(ofstream& file)
{
	file << combo->get_entry()->get_text() << endl;
	
	for(list<string>::iterator it = comboStrings.begin(); it != comboStrings.end(); ++it)
		file << *it << endl;
	
	file << "!!!COMBO DONE!!!" << endl;
}//void CF_Combo::WriteState(ofstream& file)

void CF_Combo::ReadState(ifstream& file)
{
	string input;
	
	getline(file, input);
	string box = input;
	
	getline(file, input);
	while(input != "!!!COMBO DONE!!!")
	{
		SetEntry(input, false);
		getline(file, input);
	}//while(input != "!!!COMBO DONE!!!")
		
	SetEntry(box, false);
}//void CF_Combo::ReadState(ifstream& file)

void CF_Combo::SetEntry(string str, bool activate)
{
	if(!str.empty())
	{
		for(list<string>::iterator it = comboStrings.begin(); it != comboStrings.end(); ++it)
		{
			if(*it == str)
			{		
				combo->get_entry()->set_text(str);
				if(activate)
					Activate();
				return;
			}//if(*it == str)
		}//for(list<string>::iterator it = comboStrings.begin(); it != comboStrings.end(); ++it)

		comboStrings.push_front(str);
		GList* cList = NULL;
		for(list<string>::iterator it = comboStrings.begin(); it != comboStrings.end(); ++it)
			cList = g_list_append(cList, const_cast<char*>((*it).c_str()));
		combo->set_popdown_strings(cList);
		g_list_free(cList);
	}//if(!str.empty())
	
	combo->get_entry()->set_text(str);
	if(activate) 
		Activate();
}//void CF_Combo::SetEntry(const string& str)

void CF_Combo::ClearList()
{
	comboStrings.clear();
	GList* cList = NULL;
	char* nothing = "";
	cList = g_list_append(cList, static_cast<void*>(&nothing));
	combo->set_popdown_strings(cList);
	g_list_free(cList);
}//void CF_Combo::ClearList()

void CF_Combo::AddToList(list<string>& add)
{
	//save current box contents
	string initial = combo->get_entry()->get_text();
	
	//set up insertion point
	list<string>::iterator insertIt = comboStrings.begin();
	
	//make sure strings are unique and add to list
	for(list<string>::iterator addIt = add.begin(); addIt != add.end(); ++addIt)
	{
		bool found = false;
		for(list<string>::iterator uIt = comboStrings.begin(); (uIt != comboStrings.end()) && !found; ++uIt)
		{
			if((*uIt) == (*addIt))
				found = true;
		}//for(list<string>::iterator uIt = comboStrings.begin(); uIt != comboStrings.end(); ++uIt)
		
		if(!found)
			comboStrings.insert(insertIt, *addIt);
	}//for(list<string>::iterator addIt = add.begin(); addIt != add.end(); ++addIt)
	
	//update combo box list
	GList* cList = NULL;
	for(list<string>::iterator it = comboStrings.begin(); it != comboStrings.end(); ++it)
		cList = g_list_append(cList, const_cast<char*>((*it).c_str()));
	combo->set_popdown_strings(cList);
	g_list_free(cList);	
	
	//put initial contents back in box
	combo->get_entry()->set_text(initial);
}//void CF_Combo::AddToList(list<string>& add)


//END CF_Combo



//////////////////////////////////////////////////
//BEGIN GLOBALS


//END GLOBALS


