/*****************************************************************************/
//  cf_toolbarwidget.hh
//  Copyright (C) 1999 Brett Hall <swizin@rain.org>
//  
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or 
//  (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 USA 
/*****************************************************************************/


//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// cf_toolbarwidget.hh
//	-Base class for widgets packed onto toolbars
//
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


#ifndef CF_TOOLBARWIDGET_HH
#define CF_TOOLBARWIDGET_HH

#include <string>
#include <list>
#include <vector>
#include <map>
#include <fstream>
#include <gtk--/widget.h>
#include <gtk--/button.h>
#include <gtk--/toolbar.h>
#include <gtk--/pixmap.h>
#include <gtk--/window.h>
#include <gtk--/togglebutton.h>
#include <gtk--/entry.h>
#include <gtk--/combo.h>
#include <gtk--/tooltips.h>
#include "cf_scriptvar.hh"
#include "cf_command.hh"
#include "cf_callback.hh"

using namespace std;


//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//CF_ToolbarWidget

/**Base class for toolbar widgets.  Each toolbar widget has a type retrievable using GetType()
	 and a name, user GetName(). Widgets can also read/write their state through the
	  Write/ReadState() methods.
*/
class CF_ToolbarWidget
{
public:
	///Type of widget.  Four possibilities: BUTTON, TOGGLE, TEXT, COMBO
	enum Type {BUTTON, TOGGLE, TEXT, COMBO};
	
	/**Base class initailizer
		@param nameIn widget name
		@param typeIn widget type
	*/
	CF_ToolbarWidget() : type(BUTTON) {}
	
	virtual ~CF_ToolbarWidget() {}
	
	///Returns name of widget
	const string& GetName() {return name;}
	
	///Returns type of widget
	Type GetType()	{return type;}
	
	///Returns the gtk-- widget contained in object
	virtual Gtk_Widget& GetWidget() = 0;
	
	///Writes state of widget to file
	virtual void WriteState(ofstream& file) = 0;
	
	///Reads state of widget from file
	virtual void ReadState(ifstream& file) = 0;
	
	///Sets the built-in command manager used by all widgets
	static void SetCommandMgr(CF_CommandMgr* mgr) {functions = mgr;}
	
	///Sets the script varianble manager used by all widgets
	static void SetVarMgr(CF_VarMgr* mgr) {varMgr = mgr;}
	
	///Sets the icon path used by all widgets
	static void SetIconPath(string path);
	
protected:	
	void TokenizeInitStr(const string& initStr, vector<string>& tokens); //splits inialization string into tokens
	void VerifyMatching(const string& str);	//verifies that all {} and () are matched
	string FindIcon(const string& iconName);	//searches icon paths for icon file
	
	string name;
	Type type;
	static list<string> iconPaths;
	static CF_CommandMgr* functions;
	static CF_VarMgr* varMgr;
};

//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!///
// class CF_Button

class CF_Button : public CF_ToolbarWidget
{
public:
	/**Creates a button and appends it to the given toolbar
		@param initStr String from script file that creates button
		@param bar Toolbar to append button to
		@param win Window that toolbar will go in
	*/
	CF_Button(const string& initStr, Gtk_Toolbar& bar, Gtk_Window& win);	
	
	
	~CF_Button();
	
	///Connected to button's "clicked" signal
	void Clicked();
	
	Gtk_Widget& GetWidget()	{return *(reinterpret_cast<Gtk_Widget*>(button));}
	void WriteState(ofstream& file) {}
	void ReadState(ifstream& file) {}

private:
	CEList cmds;
	Gtk_Button* button;
	Gtk_Pixmap* icon;
};//class CF_Button


//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!///
// class CF_Toggle

class CF_Toggle : public CF_ToolbarWidget
{
public:
	/**Creates a toggle button and appends to toolbar
		@param initStr String from script file that creates button
		@param bar Toolbar to append button to
		@param win WIndow that toolbar will go in
	*/
	CF_Toggle(const string& initStr, Gtk_Toolbar& bar, Gtk_Window& win);
	
	~CF_Toggle();
	
	///Connected to buttons "toggled" signal
	void Toggle();
	
	///Sets the toggle's state
	void SetState(const string& state);
	
	Gtk_Widget& GetWidget()	{return *(reinterpret_cast<Gtk_Widget*>(button));}
	void WriteState(ofstream& file);
	void ReadState(ifstream& file);

private:
	Gtk_ToggleButton* button;
	Gtk_Pixmap* icon;
	CEList upCmd;
	CEList downCmd;
	CF_ScriptVar* stateVar;
	bool disableCmd;
};//class CF_Toggle

//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!///
// class CF_Text

class CF_Text : public CF_ToolbarWidget
{
public:
	/**Creates a text entry and appends to toolbar
		@param initStr String from script file that creates entry
		@param bar Toolbar to append entry to
	*/
	CF_Text(const string& initStr, Gtk_Toolbar& bar);
	
	~CF_Text();
	
	///Connected to entry's "changed" signal
	void Changed();
	///Connected to entry's "activate" signal
	void Activate();
	
	///Sets the entry box's contents
	void SetEntry(const string& str, bool activate)	{entry->set_text(str); if(activate) Activate();}
	
	Gtk_Widget& GetWidget()	{return *(reinterpret_cast<Gtk_Widget*>(entry));}
	void WriteState(ofstream& file);
	void ReadState(ifstream& file);

private:
	Gtk_Entry* entry;
	CF_ScriptVar* stateVar;
	CEList cmds;
};//class CF_Text

//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!///
// class CF_Combo

class CF_Combo : public CF_ToolbarWidget
{
public:
	/**Creates a combo and appends to toolbar
		@param initStr String from script file that creates combo box
		@param bar Toolbar to append combo box to
	*/
	CF_Combo(const string& initStr, Gtk_Toolbar& bar);
	
	~CF_Combo();
	
	///Connected to entry's "changed" signal
	void Changed();
	///Connected to entry's "activate" signal
	void Activate();
	
	///sets the entry box string and adds to list if not already there (unless string is empty)
	void SetEntry(string str, bool activate);	
	
	///Clears the combo's list without touching entry box
	void ClearList();
	
	///Prepends strings to list without touching entry box
	void AddToList(list<string>& add);
	
	Gtk_Widget& GetWidget()	{return *(reinterpret_cast<Gtk_Widget*>(combo));}
	void WriteState(ofstream& file);
	void ReadState(ifstream& file);
	
private:	
	Gtk_Combo* combo;
	Gtk_Tooltips* tip;
	list<string> comboStrings;
	CF_ScriptVar* stateVar;
	CEList cmds;
};//class CF_Combo


#endif
