/*****************************************************************************/
//  cf_widget.cc
//  Copyright (C) 1999 Brett Hall <swizin@rain.org>
//  
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or 
//  (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 USA 
/*****************************************************************************/


//////////////////////////////////////////////////
// cf_widget.cc
//
//////////////////////////////////////////////////


#include "cf_widget.hh"
#include "cf_exception.hh"
#include "pathinfo.hh"
#include "appinfo.hh"
#include <gtk--/handlebox.h>
#include <fstream>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>


//////////////////////////////////////////////////
//BEGIN CF_Widget

class CF_WidgetOptions		//used when parsing options block in script file
{
public:
	CF_WidgetOptions();
	
	string orientation;
	string iconPath;
	string startup;
	string shutdown;
	string saveState;
	string saveVars;
};//class CF_WidgetOptions

CF_WidgetOptions::CF_WidgetOptions()
{
	orientation = "HORZ";
}//CF_OptionsWidget::CF_OptionsWidget()

CF_Widget::CF_Widget(const string& fileName, Gtk_Window& win, CF_CommandMgr* cmdMgr, CF_OptionMgr* optionsMgr) :
	clearBox_bi(widgets), clearList_bi(widgets), addToList_bi(widgets), setBox_bi(widgets),
	setToggle_bi(widgets), setVar_bi(vars), getFilename_bi(vars, widgets)
{
	//set up built-in command mgr
	functions = cmdMgr;
	//if cmd mgr is NULL then create one
	if(functions == NULL)
	{
		functions = new CF_CommandMgr;
		CF_Assert(functions != NULL, "CF_Widget::CF_Widget()");
		delFunctions = true;
	}//if(function == NULL)
	else
		delFunctions = false;
	//register the widget's built in commands
	functions->RegisterFunction(&echo);
	functions->RegisterFunction(&clearBox_bi);
	functions->RegisterFunction(&clearList_bi);
	functions->RegisterFunction(&addToList_bi);
	functions->RegisterFunction(&setBox_bi);
	functions->RegisterFunction(&setToggle_bi);
	functions->RegisterFunction(&shell_bi);
	functions->RegisterFunction(&setVar_bi);
	functions->RegisterFunction(&getFilename_bi);
	//set function mgr's in use flag
	functions->SetInUse();
	
	//set up options manager
	options = optionsMgr;
	//if options mgr is NULL then create one
	if(options == NULL)
	{
		options = new CF_OptionMgr;
		CF_Assert(options != NULL, "CF_Widget::CF_Widget()");
		delOptions = true;
	}//if(function == NULL)
	else
		delOptions = false;
	CF_WidgetOptions wOptions;	
	
	//make sure path in home diectory exists for state saving
	string homeDir = PathInfo::GetInstance().GetAppHomeDir();
	struct stat dirStat;
	if(stat(homeDir.c_str(), &dirStat) != 0)
	{
		//directory doesn't exist (or other error) => create it
		if(mkdir(homeDir.c_str(), 0770) != 0)
			cerr << "Couldn't create " << homeDir << ", won't be able to save state" << endl;
	}//if(stat(homeDir.c_str(), &dirStat) != 0)
	else
	{
		//make sure it is a directory
		if(!S_ISDIR(dirStat.st_mode))
			cerr << homeDir << "isn't a directory, won't be able to save state" << endl;
	}//else

	//open file
	string pathName = PathInfo::GetInstance().BuildPathName(fileName);
	if(pathName.empty())
		throw CF_Exception("CF_Widget::CF_Widget()", string("Couldn't open file ") + fileName);
	ifstream file(pathName.c_str());
	if(!file)
		throw CF_Exception("CF_Widget::CF_Widget()", string("Couldn't open file ") + fileName);
		
	//first pass: parse option and var blocks
	int lineNum = 0;
	try
	{
		enum {GLOBAL, OPTIONS, VARS, TOOLBAR} state = GLOBAL;
		while(!file.eof())
		{
			//read a line
			string line = ReadLine(file, lineNum);
			if(line.empty())
				continue;
				
			//interpret line according to state
			switch(state)
			{
			case GLOBAL:
				if(line == "BEGIN OPTIONS")
					state = OPTIONS;
				else if(line == "BEGIN VARS")
					state = VARS;
				else if((line == "BEGIN TOOLBAR:YES") || (line == "BEGIN TOOLBAR:NO"))
					state = TOOLBAR;
				else if(line == "NEW ROW") {}
				else
					throw CF_ParseException("CF_Widget::CF_Widget()", "Unrecognized command");
				break;

			case OPTIONS:
				if((line == "BEGIN OPTIONS") || (line == "BEGIN VARS") ||
					 (line == "BEGIN TOOLBAR:YES") || (line == "BEGIN TOOLBAR:NO") || (line == "NEW ROW"))
					 throw CF_ParseException("CF_Widget::CF_Widget()", "New block started in block");
				if(line == "END OPTIONS")
					state = GLOBAL;
				else
				{
					//split line into option and arg
					size_t pos = line.find('=');
					if(pos == string::npos)
						throw CF_ParseException("CF_Widget::CF_Widget()", "Bad option line");
					string option = line.substr(0, pos);
					string arg = line.substr(pos + 1, string::npos);
					if(option == "ORIENTATION")
						wOptions.orientation = arg;
					else if(option == "ICON_PATH")
						wOptions.iconPath = arg;
					else if(option == "START_UP")
						wOptions.startup = arg;
					else if(option == "SHUTDOWN")
						wOptions.shutdown = arg;
					else if(option == "SAVE_STATE")
						wOptions.saveState = arg;
					else if(option == "SAVE_VARS")
						wOptions.saveVars = arg;
					else
						if(!options->SetOptionFound(option, arg))
							cerr << "Unknown Option: " + option + " at " << fileName << ":" << lineNum << endl;
				}
				break;

			case VARS:
				if((line == "BEGIN OPTIONS") || (line == "BEGIN VARS") ||
					 (line == "BEGIN TOOLBAR:YES") || (line == "BEGIN TOOLBAR:NO") || (line == "NEW ROW"))
					 throw CF_ParseException("CF_Widget::CF_Widget()", "New block started in block");
				if(line == "END VARS")
					state = GLOBAL;
				else
				{
					//split line into var and initValue
					size_t pos = line.find('=');
					if(pos == string::npos)
						throw CF_ParseException("CF_Widget::CF_Widget()", "Bad option line");
					string var = line.substr(0, pos);
					string init = line.substr(pos + 1, string::npos);
					if(!vars.AddVar(var, init))
						throw CF_ParseException("CF_Widget::CF_Widget()", "Variable name collision");
				}
				break;

			case TOOLBAR:
				if((line == "BEGIN OPTIONS") || (line == "BEGIN VARS") ||
					 (line == "BEGIN TOOLBAR:YES") || (line == "BEGIN TOOLBAR:NO") || (line == "NEW ROW"))
					 throw CF_ParseException("CF_Widget::CF_Widget()", "New block started in block");
				if(line == "END TOOLBAR")
					state = GLOBAL;
				break;
			}//switch(state)
		}//while(!file.eof())
		
		//second pass: create toolbars
		
		//re-init file state
		file.seekg(0);
		file.clear();
		if(!file)
			throw CF_Exception("CF_Widget::CF_Widget()", "Couldn't clear file state for second pass");
			
		//set up toolbar widgets
		CF_ToolbarWidget::SetCommandMgr(functions);
		CF_ToolbarWidget::SetVarMgr(&vars);
		if(!wOptions.iconPath.empty())
			CF_ToolbarWidget::SetIconPath(wOptions.iconPath);
			
		//set up boxes
		bool vert = (wOptions.orientation == "VERT");
		if(vert)
			box = reinterpret_cast<Gtk_Box*>(new Gtk_HBox);
		else
			box = reinterpret_cast<Gtk_Box*>(new Gtk_VBox);
		CF_Assert(box != NULL, "CF_Widget::CF_Widgeet()");
		//box->show();
		Gtk_Box* boxPtr = reinterpret_cast<Gtk_Box*>(box);
					
		//start reading lines again
		state = GLOBAL;
		lineNum = 0;
		Gtk_Toolbar* curBar = NULL;
		Gtk_Box* curBox;
		if(vert)
			curBox = reinterpret_cast<Gtk_Box*>(new Gtk_VBox);
		else
			curBox = reinterpret_cast<Gtk_Box*>(new Gtk_HBox);
		CF_Assert(curBox != NULL, "CF_Widget::CF_Widgeet()");
		//curBox->show();
		boxes.push_back(curBox);
		boxPtr->pack_start(*curBox);
		while(!file.eof())
		{
			string line = ReadLine(file, lineNum);
			if(line.empty())
				continue;
				
			switch(state)
			{
			case GLOBAL:
				if(line == "BEGIN OPTIONS")
					state = OPTIONS;
				else if(line == "BEGIN VARS")
					state = VARS;
				else if((line == "BEGIN TOOLBAR:YES") || (line == "BEGIN TOOLBAR:NO"))
				{
					state = TOOLBAR;
					
					//find ":"
					size_t pos = line.find(':');
					bool useHandle = (line.substr(pos  + 1, line.length() - pos) == "YES");
					Gtk_HandleBox* handle = NULL;
					if(useHandle)
					{
						handle = new Gtk_HandleBox;
						CF_Assert(handle != NULL, "CF_Widget::CF_WIdget()");
						//handle->show();
						toolbars.push_back(handle);
					}//if(line.substr(pos  + 1, line.length() - 1) == "YES")
					if(vert)
						curBar = new Gtk_Toolbar(GTK_ORIENTATION_VERTICAL);
					else
						curBar = new Gtk_Toolbar;
					CF_Assert(curBar != NULL, "CF_Widget::CF_WIdget()");
					//curBar->show();
					toolbars.push_back(curBar);
					if(useHandle)
					{
						handle->add(*curBar);
						curBox->pack_start(*handle);
					}//if(useHandle)
					else
						curBox->pack_start(*curBar);
				}//else if((line == "BEGIN TOOLBAR:YES") || (line == "BEGIN TOOLBAR:NO"))
				else if(line == "NEW ROW")
				{
					if(vert)
						curBox = reinterpret_cast<Gtk_Box*>(new Gtk_VBox);
					else
						curBox = reinterpret_cast<Gtk_Box*>(new Gtk_HBox);
					CF_Assert(curBox != NULL, "CF_Widget::CF_Widget()");
					//curBox->show();
					boxes.push_back(curBox);
					boxPtr->pack_start(*curBox);
				}//else if(line == "NEW ROW")
				else
					throw CF_ParseException("CF_Widget::CF_Widget()", "Unrecognized command");
				break;
				
			case OPTIONS:
				if(line == "END OPTIONS")
					state = GLOBAL;
				break;
				
			case VARS:
				if(line == "END VARS")
					state = GLOBAL;
				break;
				
			case TOOLBAR:
				if(line == "END TOOLBAR")
					state = GLOBAL;
				else
				{
					//split into widget type and init line
					size_t pos = line.find(':');
					if(pos == string::npos)
						throw CF_ParseException("CF_Widget::CF_Widget()", "Bad widget line");
					string type = line.substr(0, pos);
					string init = line.substr(pos + 1, line.length() - pos);
					
					//create widget
					if(type == "BUTTON")
					{
						CF_Button* button = new CF_Button(init, *curBar, win);
						CF_Assert(button != NULL, "CF_Widget::CF_Widget()");
						AddWidget(button);
					}//if(type == "BUTTON")
					else if(type == "TOGGLE")
					{
						CF_Toggle* toggle = new CF_Toggle(init, *curBar, win);
						CF_Assert(toggle != NULL, "CF_Widget::CF_Widget()");
						AddWidget(toggle);		
					}//else if(type == "TOGGLE")
					else if(type == "TEXT")
					{
						CF_Text* text = new CF_Text(init, *curBar);
						CF_Assert(text != NULL, "CF_Widget::CF_Widget()");
						AddWidget(text);				
					}//else if(type == "TEXT")
					else if(type == "COMBO")
					{
						CF_Combo* combo = new CF_Combo(init, *curBar);
						CF_Assert(combo != NULL, "CF_Widget::CF_Widget()");
						AddWidget(combo);						
					}//else if(type == "COMBO")
					else if(type == "SEPARATOR")
					{
						curBar->append_space();
						curBar->set_space_size(atoi(init.c_str()));		
					}//else if(type == "SEPARATOR")
					else
						throw CF_ParseException("CF_Widget::CF_Widget()", string("Unknown widget type: ") + type);
				}//else
			}//switch(state)
		}//while(!file.eof())
		
		file.close();
	}//try
	catch(CF_ParseException& exc)
	{
		exc.DisplayMsg(fileName, lineNum, 1);
	}//catch(CF_ParseException& exc)
	
	//record save state info and restore state
	scriptName = fileName;
	if(!wOptions.saveState.empty())
	{
		if(wOptions.saveState == "ALL")
		{
			//iterarte over all widgets and save names
			for(map<const string, CF_ToolbarWidget*>::iterator it = widgets.begin(); it != widgets.end(); ++it)
			{
				if((*it).second != NULL)
					save.push_back((*it).first);
			}//for(map<const string, CF_ToolbarWidget*>::iterator it = widgets.begin(); it != widgets.end(); ++it)
		}//if(wOptions.saveState == "ALL")
		else
		{
			string curName;
			for(size_t pos = 0; pos < wOptions.saveState.length(); ++pos)
			{
				switch(wOptions.saveState[pos])
				{
				case '|':
					save.push_back(curName);
					curName.erase();
					break;

				default:
					curName += wOptions.saveState[pos];
					break;
				}//switch(wOptions.saveState[pos])
			}//for(size_t pos = 0; pos < wOptions.saveState.length(); ++pos)
			save.push_back(curName);			
		}//else
		
		//restore state
		string home = PathInfo::GetInstance().GetAppHomeDir();
		home += "%F.state";
		home.replace(home.find("%F"), 2, fileName);
		ifstream file(home.c_str());
		if(!file)
			cerr << "Couldn't read state file: " << home << endl;
		else
		{
			while(!file.eof())
			{
				string line;
				getline(file, line);
				
				if(!line.empty())
				{
					CF_ToolbarWidget* widget = widgets[line];
					if(widget == NULL)
					{
						cerr << "Unknown widget " << line << " in state file" << endl;
					}//if(widget == NULL)
					else
					{
						widget->ReadState(file);
					}//else					
				}//if(!line.empty())
			}//while(!file.eof())
			
			file.close();	
		}//else
	}//if(!wOptions.saveState.empty())
	
	if(!wOptions.saveVars.empty())
	{
		if(wOptions.saveVars == "ALL")
		{
			//iterate over all vars and save names
			for(CF_VarMgr::iterator it = vars.begin(); it != vars.end(); ++it)
			{
				if((*it).second != NULL)
					saveVars.push_back((*it).first);
			}//for(map<const string, CF_ToolbarWidget*>::iterator it = widgets.begin(); it != widgets.end(); ++it)
		}//if(wOptions.saveState == "ALL")
		else
		{
			string curName;
			for(size_t pos = 0; pos < wOptions.saveVars.length(); ++pos)
			{
				switch(wOptions.saveVars[pos])
				{
				case '|':
					saveVars.push_back(curName);
					curName.erase();
					break;

				default:
					curName += wOptions.saveVars[pos];
					break;
				}//switch(wOptions.saveState[pos])
			}//for(size_t pos = 0; pos < wOptions.saveState.length(); ++pos)
			saveVars.push_back(curName);			
		}//else
		
		//restore state
		string home = PathInfo::GetInstance().GetAppHomeDir();
		home += "%F.var";
		home.replace(home.find("%F"), 2, fileName);
		ifstream file(home.c_str());
		if(!file)
			cerr << "Couldn't read state file: " << home << endl;
		else
		{
			while(!file.eof())
			{
				string line;
				getline(file, line);
				if(!line.empty())
				{
					size_t pos = line.find('=');
					string name = line.substr(0, pos);
					string value = line.substr(pos + 1, line.length() - pos);
					CF_ScriptVar* var = vars.GetVar(name);
					if(var == NULL)
					{
						cerr << "Unknown var " << name << " in state file" << endl;
					}//if(widget == NULL)
					else
					{
						var->SetValue(value);
					}//else					
				}//if(!line.empty())
			}//while(!file.eof())
			
			file.close();	
		}//else
	}//if(!wOptions.saveVars.empty())
	
	//set up startup/shutdown commands
	if(!wOptions.startup.empty())
		CF_CommandElement::CreateCommands(wOptions.startup, startupCmds, functions, &vars);
	if(!wOptions.shutdown.empty())
		CF_CommandElement::CreateCommands(wOptions.shutdown, shutdownCmds, functions, &vars);
		
	//execute start up commands
	startupCmds.ExpandCommands();
	
	//show the box
	box->show_all();
}//CF_Widget::CF_Widget(const string& fileName, CF_CommandMgr* cmdMgr, CF_OptionsMgr* optionsMgr)


//CF_Widget::~CF_Widget: deletes all widgets, tool bars and script vars
//returns: <nothing>
CF_Widget::~CF_Widget()
{
	//execute shutdown commands
	shutdownCmds.ExpandCommands();
	
	//save state of widgets
	string home = PathInfo::GetInstance().GetAppHomeDir();
	home += "%F.state";
	home.replace(home.find("%F"), 2, scriptName);
	ofstream file(home.c_str());
	if(!file)
		cerr << "Couldn't write state to file: " << home << endl;
	else
	{
		for(list<string>::iterator it = save.begin(); it != save.end(); ++it)
		{
			CF_ToolbarWidget* w = widgets[*it];
			if(w != NULL)
			{
				file << *it << endl;
				w->WriteState(file);
			}//if(w != NULL)
		}//for(list<string>::iterator it = widgets.begin(); it != widgets.end(); ++it)
		file.close();
	}//else

	//save state of vars
	home = PathInfo::GetInstance().GetAppHomeDir();
	home += "%F.var";
	home.replace(home.find("%F"), 2, scriptName);
	file.open(home.c_str());
	if(!file)
		cerr << "Couldn't write state to file: " << home << endl;
	else
	{
		for(list<string>::iterator it = saveVars.begin(); it != saveVars.end(); ++it)
		{
			CF_ScriptVar* var = vars.GetVar(*it);
			if(var != NULL)
			{
				file << *it << "=" << var->GetStringValue() << endl;
			}//if(w != NULL)
		}//for(list<string>::iterator it = widgets.begin(); it != widgets.end(); ++it)
		file.close();
	}//else

	
	//get rid of widgets
	for(map<const string, CF_ToolbarWidget*>::iterator wIt = widgets.begin(); wIt != widgets.end(); ++wIt)
	{
		if((*wIt).second != NULL)
			delete (*wIt).second;
	}//for(map<const string, CF_ScriptVar*>::iterator wIt = widgets.begin(); wIt != widgets.end(); ++wIt)
	
	//get rid of toolbars
	while(!toolbars.empty())
	{
		if(toolbars.front() != NULL)
			delete toolbars.front();
		toolbars.pop_front();
	}//while(!toolbars.empty())
	
	//get rid of boxes
	while(!boxes.empty())
	{
		if(boxes.front() != NULL)
			delete boxes.front();
		boxes.pop_front();
	}//while(!toolbars.empty())
	if(box != NULL)
		delete box;
		
	//get rid of options and functions is nescessary
	if(delFunctions)
		delete functions;
	if(delOptions)
		delete options;
}//CF_Widget::~CF_Widget()


//CF_Widget::AddWidget: adds a widget to the widget list , does not pack widget
//	widget: widget to add
//returns: true if no name collision
bool CF_Widget::AddWidget(CF_ToolbarWidget* widget)
{
	if(widgets[widget->GetName()] != NULL)
		return false;
	
	widgets[widget->GetName()] = widget;
	return true;
}//bool CF_Widget::AddWidget(CF_ToolbarWidget* widget)

string CF_Widget::ReadLine(ifstream& file, int& lineNum)
{
	//string to return
	string modLine;
	
	//read from file until no more concatenated lines
	bool done = false;
	while(!done)
	{
		//make sure we haven't hit eof
		if(file.eof())
			throw CF_Exception("CF_Widget::ReadLine()", "Unexpected end of file");
			
		//read a line
		string line;
		getline(file, line);
		
		//remove comment and leading and trailing whitespace
		size_t posComment = line.find('#');
		if(posComment != string::npos)
			line.erase(posComment, line.length() - posComment);
		size_t posLead = line.find_first_not_of(" \t");
		if(posLead != string::npos)
			line.erase(0, posLead);
		else
		{
			++lineNum;
			return modLine;	//nothing on line but whitespace
		}//else 
		size_t posTrail = line.find_last_not_of(" \t");
		if(posTrail != string::npos)
			line.erase(posTrail + 1, line.length() - posTrail);
			
		//append line to modLine while stripping extraneous whitespace
		bool inWhiteSpace = false;
		bool prevControl = false;
		char prevChar = ' ';
		string whiteSpace;
		for(size_t pos = 0; pos < line.length(); ++pos)
		{
			switch(line[pos])
			{
			case ' ':
			case '\t':
				inWhiteSpace = true;
				whiteSpace += line[pos];
				break;
				
			case ',':
			case ';':
			case ':':
			case '?':
			case '|':
			case '!':
			case '=':
			case '(':
			case ')':
				if(prevChar != '\\')
				{
					//not escaped
					inWhiteSpace = false;
					prevControl = true;
					whiteSpace.erase();
					modLine += line[pos];
				}//if(prevChar != '\\')
				else
				{
					//escaped, treat like regular char
					if(inWhiteSpace)
					{
						inWhiteSpace = false;
						if(!prevControl)
							modLine += whiteSpace;
						whiteSpace.erase();
					}//if(inWhiteSpace)
					prevControl = false;
					modLine += line[pos];
				}//else
				break;
				
			case '{':
				if(inWhiteSpace)
				{
					inWhiteSpace = false;
					if(!prevControl)
						modLine += whiteSpace;
					whiteSpace.erase();
				}//if(inWhiteSpace)
				prevControl = (prevChar != '\\');
				modLine += line[pos];
				break;
				
			case '}':
				if(prevChar != '\\')
				{
					inWhiteSpace = false;
					whiteSpace.erase();
				}//if(prevChar != '\\')
				else
				{
					if(inWhiteSpace)
					{
						inWhiteSpace = false;
						if(!prevControl)
							modLine += whiteSpace;
						whiteSpace.erase();
					}//if(inWhiteSpace)	
				}//else
				modLine += line[pos];
				prevControl = false;
				break;
				
			default:
				if(inWhiteSpace)
				{
					inWhiteSpace = false;
					if(!prevControl)
						modLine += whiteSpace;
					whiteSpace.erase();
				}//if(inWhiteSpace)
				prevControl = false;
				modLine += line[pos];
				break;
			}//switch(line[pos])
			
			prevChar = line[pos];
		}//for(size_t pos = 0; pos < line.length(); ++pos)
		
		if(modLine[modLine.length() - 1] == '\\')
			modLine.erase(modLine.length() - 1, 1);
		else
			done = true;
		++lineNum;
	}//while(!done)
	
	return modLine;
}//string ReadLine(ifsteam& file)


//CF_Widget::GetWidget: gets a widget from the widget list
//	name: name of widget to get
//returns: widget
CF_ToolbarWidget* CF_Widget::GetWidget(const string& name)
{
	return widgets[name];		
}//CF_ToolbarWidget* CF_Widget::GetWidget(const string& name)


//END CF_Widget


/////////////////////////////////////////////////////
//BEGIN CF_Widget::CF_EchoCmd

string CF_Widget::CF_EchoCmd::Call(vector<string>& args)
{
	for(vector<string>::iterator it = args.begin(); it != args.end(); ++it)
		cout << *it << endl;
		
	return "";
}//string CF_Widget::CF_EchoCmd::Call(vector<string>& args)


//END CF_Widget::CF_EchoCmd


//////////////////////////////////////////////////
//BEGIN GLOBALS


//END GLOBALS


